#include <stdlib.h>
#include <vga.h>
#include <vgagl.h>
#include "math.h"

//see vga_setmode
#define MYWIDTH  320
#define MYHEIGHT 240
//#define GMODE G320x240x256
#define GMODE G320x240x256

#define BOUND 35
#define THRESH 20
#define MAXCOUNT 2000000
#define POINT_BUF_LENGTH 20

float lastxpixel, xpixel, lastypixel, ypixel;
 
float INITIAL_XMAX=4.0, INITIAL_XMIN=-4.0, INITIAL_YMAX=3.0, INITIAL_YMIN=-3.0;
float MOVE_FRACTION_X = 0.05;
float MOVE_FRACTION_Y = 0.05;
float SCREEN_FRACTION_X = 0.1;
float SCREEN_FRACTION_Y = 0.1;

char rgbbuf[MYWIDTH * MYHEIGHT];

float xmax_queue[POINT_BUF_LENGTH];
float xmin_queue[POINT_BUF_LENGTH];
float ymax_queue[POINT_BUF_LENGTH];
float ymin_queue[POINT_BUF_LENGTH];
int points;

void init(void);
void plot_picture(void);
void paint_picture(void);
void update(void);
int count(int, int);
int iscomplex(int);


float xmin, xmax, ymin, ymax, t_xmin, t_xmax, t_ymin, t_ymax;
float p_xmin, p_xmax, p_ymin, p_ymax;

GraphicsContext *physicalscreen;
GraphicsContext *virtualscreen;

int main(void) {
  vga_init();
  vga_setmode(GMODE);
  gl_setcontextvga(GMODE);
  physicalscreen = gl_allocatecontext();
  gl_getcontext(physicalscreen);
  gl_setcontextvgavirtual(GMODE);
  virtualscreen = gl_allocatecontext();
  gl_getcontext(virtualscreen);
  gl_setcontext(virtualscreen);

  srandom(time(NULL));

  init();

  while (!vga_getkey()) {
    plot_picture();
    paint_picture();
    update();
    //usleep(100000);
  }

  vga_setmode(TEXT);
  return 0;
}

int count(int x, int y) {
  float z=0, zi=0, z_2=0, zi_2=0, zprev, c, ci;
  int n;
  c = xmin+x*xpixel;
  ci = ymin+y*ypixel;
  //for (n=0;n<256 && z < 2;n++) { // not the "proper" but NIFTY artifacts
  for (n=0;n<64 && (z_2 + zi_2) < 4;n++) {
    zprev=z;
    z_2=z*z;
    zi_2=zi*zi;
    z = z_2 - zi_2 + c;
    zi= 2 * zprev * zi + ci;
  }
  return n;
}

void init() {
  xmin = INITIAL_XMIN; xmax = INITIAL_XMAX;
  ymin = INITIAL_YMIN; ymax = INITIAL_YMAX;
  t_xmax=xmax;
  t_xmin=xmin;
  t_ymax=ymax;
  t_ymin=ymin;
  p_xmax=xmax*2;
  p_xmin=xmin*2;
  p_ymax=ymax*2;
  p_ymin=ymin*2;
}

void plot_picture() {
  int x, y, temp;
  char *pos;
  pos = rgbbuf;
  for (y = 0; y < MYHEIGHT; y++) {
    for (x = 0; x < MYWIDTH; x++) {
      temp=count(x,y);
      *pos++ = temp;
    }
  }
}

int iscomplex(int point) {
  int x = point % MYWIDTH;
  int y = point / MYWIDTH;
  int up,left,right,down,upleft,upright,downleft,downright,center;
  center = rgbbuf[(x+(MYWIDTH*y))];
  up = rgbbuf[(x+(MYWIDTH*(y+1)))];
  down = rgbbuf[(x+(MYWIDTH*(y-1)))];
  left = rgbbuf[(x+(MYWIDTH*y))-1];
  right = rgbbuf[(x+(MYWIDTH*y))+1];
  if ((center > THRESH) && (up != down) && (up != center) && (down != center) && (left != right) && (right != center)) return 1;
  else return 0;
}

void paint_picture(){
  int x,y,c;
  for (y=0;y<MYHEIGHT;y++) {
    for (x=0;x<MYWIDTH;x++) {
      c=rgbbuf[x+MYWIDTH*y];
      gl_setpixelrgb(x,y,c*2,(c==0)?0:64+c*3,(c==0)?0:128+c*2); // x, y, r, g, b
      //gl_setpixelrgb(x,y,c*4,c*4,c*4); // x, y, r, g, b
      //gl_setpixelrgb(x,y,c,c,c); // x, y, r, g, b
      //gl_setpixelrgb(x,y,c*4,(c==0)?0:256-c*4,c); // x, y, r, g, b
    }
  }
  gl_copyscreen(physicalscreen);
}

void update() {
  float xbound, ybound;
  static float lastxpixel, lastypixel;
  static int outercount=0;
  float point1x=0, point1y=0, point2x=0, point2y=0;
  float width,height;
  int x, y, point, mycount=0;
  int forcePop=0;
  int forceNoPop=0;
  width=xmax-xmin;
  height=ymax-ymin;
  xpixel = width/WIDTH;
  ypixel = height/HEIGHT;
  xbound = fabs(xpixel) * BOUND;
  ybound = fabs(ypixel) * BOUND;


  if ( lastxpixel == xpixel || lastypixel == ypixel //case 1 - a pixel is 0.0f wide, we're not going to get anywhere. fast.
       //|| (t_xmax == t_xmin || t_ymax == t_ymin)//case 2 - our target is too small
       //|| p_xmin == xmin || p_xmax == xmax || p_ymin == ymin || p_ymax == ymax // case 3 -- we don't seem to be getting anywhere
     ) {
    forcePop=1;
  }
  if (points > 0 && (xmax_queue[points-1] == 0 && xmin_queue[points-1] == 0)) {
    forceNoPop=1;
  }
  //if we're within BOUND pixels for xmax, xmin, ymax, ymin -- pick a point
  if (forcePop || (fabs(xmax-t_xmax) < xbound && fabs(xmin-t_xmin) < xbound && fabs(ymax-t_ymax) < ybound && fabs(ymin-t_ymin) < ybound)) {
    //if we gotta, we gotta
    if (forcePop) {
printf("FORCEPOP\n");
      points=0;
      t_xmax = INITIAL_XMAX;
      t_xmin = INITIAL_XMIN;
      t_ymax = INITIAL_YMAX;
      t_ymin = INITIAL_YMIN;
    //if there is anything on the queue, 50/50 to pick off it
    } else if ((points > 0 && ((double)random()/RAND_MAX) > .5)){
printf("POP\n");
      points--;
      t_xmax = xmax_queue[points];
      t_xmin = xmin_queue[points];
      t_ymax = ymax_queue[points];
      t_ymin = ymin_queue[points];
    //take random samples until two points are above threshold
    } else {
printf("picking a new point because we're at: \n");
printf("width: %.11e, height: %.11e\n",width,height);
printf("pix[%.11e,%.11e,%.11e,%.11e]\n",lastxpixel,xpixel,lastypixel,ypixel);
printf("now[%.11e,%.11e,%.11e,%.11e]\n",xmin,xmax,ymin,ymax);
printf("tar[%.11e,%.11e,%.11e,%.11e]\n",t_xmin,t_xmax,t_ymin,t_ymax);
      while (point2y == 0 && mycount < MAXCOUNT) {
        point = (int) fabs(((double)random()/RAND_MAX) * WIDTH * HEIGHT);
        mycount++;
        if (iscomplex(point)) {
          x = point % WIDTH;
          y = point / WIDTH;
          if (point1y == 0) {
            point1x = xmin + x * xpixel;
            point1y = ymin + y * ypixel;
          } else {
            point2x = xmin + x * xpixel;
            point2y = ymin + y * ypixel;
          }
        }
      }
      //set target bounds to bound that point as center
      //using SCREEN_FRACTION_X/Y
      if (mycount < MAXCOUNT) {
        t_xmax = point1x + width*(SCREEN_FRACTION_X/2);
        t_xmin = point1x - width*(SCREEN_FRACTION_X/2);
        t_ymax = point1y + height*(SCREEN_FRACTION_Y/2);
        t_ymin = point1y - height*(SCREEN_FRACTION_Y/2);
        xmax_queue[points] = point2x + width*(SCREEN_FRACTION_X/2);
        xmin_queue[points] = point2x - width*(SCREEN_FRACTION_X/2);
        ymax_queue[points] = point2y + height*(SCREEN_FRACTION_Y/2);
        ymin_queue[points] = point2y - height*(SCREEN_FRACTION_Y/2);
        points++;
      } else {
        //we're not able to find anything interesting, let's start over...
        points=0;
        t_xmax = INITIAL_XMAX;
        t_xmin = INITIAL_XMIN;
        t_ymax = INITIAL_YMAX;
        t_ymin = INITIAL_YMIN;
      }
printf("tar[%.11e,%.11e,%.11e,%.11e]\n\n",t_xmin,t_xmax,t_ymin,t_ymax);
    }
outercount=0;
  }
  //move MOVE_FRACTION_X/Y
  p_xmax = xmax; p_xmin = xmin; p_ymax = ymax; p_ymin = ymin;
  xmax += (t_xmax - xmax)* MOVE_FRACTION_X;
  xmin += (t_xmin - xmin)* MOVE_FRACTION_X;
  ymax += (t_ymax - ymax)* MOVE_FRACTION_Y;
  ymin += (t_ymin - ymin)* MOVE_FRACTION_Y;
  lastxpixel=xpixel;
  lastypixel=ypixel;
  outercount++;
  if (outercount > 200) {
printf("stuck!?\n");    
printf("width: %.11e, height: %.11e\n",width,height);
printf("pix[%.11e,%.11e,%.11e,%.11e]\n",lastxpixel,xpixel,lastypixel,ypixel);
printf("now[%.11e,%.11e,%.11e,%.11e]\n\n",xmin,xmax,ymin,ymax);
  }
}
